# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


class Verbose(dict):
    def __getitem__(self, key):
        self.touched = True
        if "data" in key:
            self.data_touched = True
        return dict.__getitem__(self, key)


def test_lazy():
    form = {
        "class": "RecordArray",
        "contents": {
            "Electron": {
                "class": "ListOffsetArray64",
                "content": {
                    "class": "RecordArray",
                    "contents": {
                        "charge": {
                            "class": "NumpyArray",
                            "form_key": "load_electron_charge",
                            "format": "i",
                            "has_identities": False,
                            "inner_shape": [],
                            "itemsize": 4,
                            "parameters": {},
                            "primitive": "int64",
                        },
                        "momentum": {
                            "class": "NumpyArray",
                            "form_key": "load_electron_momentum",
                            "format": "i",
                            "has_identities": False,
                            "inner_shape": [],
                            "itemsize": 4,
                            "parameters": {},
                            "primitive": "float64",
                        },
                    },
                    "form_key": "invalid",
                    "parameters": {"__record__": "Electron"},
                },
                "form_key": "load_electron_offsets",
                "offsets": "i64",
            },
            "Muon": {
                "class": "ListOffsetArray64",
                "content": {
                    "class": "RecordArray",
                    "contents": {
                        "charge": {
                            "class": "NumpyArray",
                            "form_key": "load_muon_charge",
                            "format": "i",
                            "has_identities": False,
                            "inner_shape": [],
                            "itemsize": 4,
                            "parameters": {},
                            "primitive": "int64",
                        },
                        "momentum": {
                            "class": "NumpyArray",
                            "form_key": "load_muon_momentum",
                            "format": "i",
                            "has_identities": False,
                            "inner_shape": [],
                            "itemsize": 4,
                            "parameters": {},
                            "primitive": "float64",
                        },
                    },
                    "form_key": "invalid",
                    "parameters": {"__record__": "Muon"},
                },
                "form_key": "load_muon_offsets",
                "offsets": "i64",
            },
        },
        "form_key": "",
        "parameters": {},
    }

    container = Verbose(
        {
            "part0-load_electron_offsets-offsets": np.array(
                [0, 3, 3, 5], dtype=np.int64
            ),
            "part0-load_muon_offsets-offsets": np.array([0, 3, 3, 5], dtype=np.int64),
            "part0-load_electron_charge-data": np.array(
                [1, 2, 3, 4, 5], dtype=np.int64
            ),
            "part0-load_electron_momentum-data": np.array(
                [1.1, 2.2, 3.3, 4.4, 5.5], dtype=np.float64
            ),
            "part0-load_muon_charge-data": np.array([1, 2, 3, 4, 5], dtype=np.int64),
            "part0-load_muon_momentum-data": np.array(
                [1.1, 2.2, 3.3, 4.4, 5.5], dtype=np.float64
            ),
            "part1-load_electron_offsets-offsets": np.array(
                [0, 3, 3, 5], dtype=np.int64
            ),
            "part1-load_muon_offsets-offsets": np.array([0, 3, 3, 5], dtype=np.int64),
            "part1-load_electron_charge-data": np.array(
                [1, 2, 3, 4, 5], dtype=np.int64
            ),
            "part1-load_electron_momentum-data": np.array(
                [1.1, 2.2, 3.3, 4.4, 5.5], dtype=np.float64
            ),
            "part1-load_muon_charge-data": np.array([1, 2, 3, 4, 5], dtype=np.int64),
            "part1-load_muon_momentum-data": np.array(
                [1.1, 2.2, 3.3, 4.4, 5.5], dtype=np.float64
            ),
        }
    )
    container.touched = False

    lazy = ak.from_buffers(form, [3, 3], container, lazy=True)
    one = ak.concatenate([lazy.Electron, lazy.Muon], axis=0)

    lazy = ak.from_buffers(form, [3, 3], container, lazy=True)
    two = ak.concatenate([lazy.Muon, lazy.Muon], axis=0)

    lazy = ak.from_buffers(form, [3, 3], container, lazy=True)
    three = ak.concatenate([lazy.Electron, lazy.Electron], axis=0)

    assert not container.touched

    assert one.tolist() == 4 * [
        [
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
        ],
        [],
        [{"charge": 4, "momentum": 4.4}, {"charge": 5, "momentum": 5.5}],
    ]
    assert two.tolist() == 4 * [
        [
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
        ],
        [],
        [{"charge": 4, "momentum": 4.4}, {"charge": 5, "momentum": 5.5}],
    ]
    assert three.tolist() == 4 * [
        [
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
        ],
        [],
        [{"charge": 4, "momentum": 4.4}, {"charge": 5, "momentum": 5.5}],
    ]

    assert container.touched

    container.data_touched = False

    lazy = ak.from_buffers(form, [3, 3], container, lazy=True)
    one = ak.concatenate([lazy.Electron, lazy.Muon], axis=1)

    lazy = ak.from_buffers(form, [3, 3], container, lazy=True)
    two = ak.concatenate([lazy.Muon, lazy.Muon], axis=1)

    lazy = ak.from_buffers(form, [3, 3], container, lazy=True)
    three = ak.concatenate([lazy.Electron, lazy.Electron], axis=1)

    assert not container.data_touched

    assert one.tolist() == 2 * [
        [
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
        ],
        [],
        [
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
        ],
    ]
    assert two.tolist() == 2 * [
        [
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
        ],
        [],
        [
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
        ],
    ]
    assert three.tolist() == 2 * [
        [
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
        ],
        [],
        [
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
        ],
    ]

    form["contents"]["Electron"]["content"]["parameters"] = {}
    form["contents"]["Muon"]["content"]["parameters"] = {}

    container.data_touched = False

    lazy = ak.from_buffers(form, [3, 3], container, lazy=True)

    one = ak.concatenate([lazy.Electron, lazy.Muon], axis=1, merge=False)

    assert not container.data_touched

    assert one.tolist() == 2 * [
        [
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
            {"charge": 1, "momentum": 1.1},
            {"charge": 2, "momentum": 2.2},
            {"charge": 3, "momentum": 3.3},
        ],
        [],
        [
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
            {"charge": 4, "momentum": 4.4},
            {"charge": 5, "momentum": 5.5},
        ],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import numbers

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()


def single(obj):
    if isinstance(obj, ak.layout.Content):
        return IrregularlyPartitionedArray([obj])
    else:
        return obj


def first(obj):
    if isinstance(obj, PartitionedArray):
        return obj.partition(0)
    else:
        return obj


def every(obj):
    if isinstance(obj, PartitionedArray):
        return obj.partitions
    else:
        return [obj]


def partition_as(sample, arrays):
    stops = sample.stops
    if isinstance(arrays, dict):
        out = {}
        for n, x in arrays.items():
            if isinstance(x, PartitionedArray):
                out[n] = x.repartition(stops)
            elif isinstance(x, ak.layout.Content) and (
                x.parameter("__array__") == "string"
                or x.parameter("__array__") == "bytestring"
            ):
                out[n] = x
            elif isinstance(x, ak.layout.Content):
                out[n] = IrregularlyPartitionedArray.toPartitioned(x, stops)
            else:
                out[n] = x
        return out
    else:
        out = []
        for x in arrays:
            if isinstance(x, PartitionedArray):
                out.append(x.repartition(stops))
            elif isinstance(x, ak.layout.Content) and (
                x.parameter("__array__") == "string"
                or x.parameter("__array__") == "bytestring"
            ):
                out.append(x)
            elif isinstance(x, ak.layout.Content):
                out.append(IrregularlyPartitionedArray.toPartitioned(x, stops))
            else:
                out.append(x)
        return out


def iterate(numpartitions, arrays):
    if isinstance(arrays, dict):
        for partitionid in range(numpartitions):
            out = {}
            for n, x in arrays.items():
                if isinstance(x, PartitionedArray):
                    out[n] = x.partition(partitionid)
                else:
                    out[n] = x
            yield out
    else:
        for partitionid in range(numpartitions):
            out = []
            for x in arrays:
                if isinstance(x, PartitionedArray):
                    out.append(x.partition(partitionid))
                else:
                    out.append(x)
            yield out


def apply(function, array):
    return IrregularlyPartitionedArray([function(x) for x in array.partitions])


class PartitionedArray(object):
    @classmethod
    def from_ext(cls, obj):
        if isinstance(obj, ak._ext.IrregularlyPartitionedArray):
            subcls = IrregularlyPartitionedArray
            out = subcls.__new__(subcls)
            out._ext = obj
            return out
        else:
            return obj

    def __repr__(self):
        return repr(self._ext)

    @property
    def numpartitions(self):
        return self._ext.numpartitions

    @property
    def partitions(self):
        return self._ext.partitions

    def partition(self, partitionid):
        return self._ext.partition(partitionid)

    def partitionid_index_at(self, at):
        return self._ext.partitionid_index_at(at)

    def repartition(self, *args, **kwargs):
        return PartitionedArray.from_ext(self._ext.repartition(*args, **kwargs))

    @property
    def stops(self):
        total_length = 0
        out = []
        for x in self.partitions:
            total_length += len(x)
            out.append(total_length)
        return out

    @property
    def lengths(self):
        return [len(x) for x in self.partitions]

    def start(self, partitionid):
        return self._ext.start(partitionid)

    def stop(self, partitionid):
        return self._ext.stop(partitionid)

    def toContent(self):
        contents = self._ext.partitions
        if len(contents) == 1:
            return contents[0]

        batch = [contents[0]]
        for x in contents[1:]:
            if batch[-1].mergeable(x, mergebool=False):
                batch.append(x)
            else:
                collapsed = batch[0].mergemany(batch[1:])
                batch = [collapsed.merge_as_union(x)]

        out = batch[0].mergemany(batch[1:])
        if isinstance(out, ak._util.uniontypes):
            out = out.simplify(mergebool=False)
        return out

    def __array__(self):
        tocat = []
        for x in self.partitions:
            y = ak.operations.convert.to_numpy(x)
            if len(y) > 0:
                tocat.append(y)

        if len(tocat) > 0:
            if any(isinstance(x, numpy.ma.MaskedArray) for x in tocat):
                return numpy.ma.concatenate(tocat)
            else:
                return ak.nplike.of(tocat).concatenate(tocat)
        else:
            return y

    ############################### Content methods

    @property
    def identities(self):
        raise NotImplementedError(ak._util.exception_suffix(__file__))

    def setidentities(self, identities=None):
        raise NotImplementedError(ak._util.exception_suffix(__file__))

    @property
    def parameters(self):
        return first(self).parameters

    def setparameter(self, param, value):
        raise NotImplementedError(ak._util.exception_suffix(__file__))

    def withparameter(self, param, value):
        return self.replace_partitions(
            [x.withparameter(param, value) for x in self.partitions]
        )

    def parameter(self, *args, **kwargs):
        return first(self).parameter(*args, **kwargs)

    def purelist_parameter(self, *args, **kwargs):
        return first(self).purelist_parameter(*args, **kwargs)

    def type(self, typestrs):
        out = None
        for x in self.partitions:
            if out is None:
                out = x.type(typestrs)
            elif out != x.type(typestrs):
                raise ValueError(
                    "inconsistent types in PartitionedArray"
                    + ak._util.exception_suffix(__file__)
                )
        return out

    @property
    def form(self):
        return first(self).form

    def __len__(self):
        return len(self._ext)

    def __getitem__(self, where):
        if not isinstance(where, (bool, np.bool_)) and isinstance(
            where, (numbers.Integral, np.integer)
        ):
            return PartitionedArray.from_ext(self._ext.getitem_at(where))

        elif isinstance(where, slice):
            return PartitionedArray.from_ext(
                self._ext.getitem_range(where.start, where.stop, where.step)
            )

        elif isinstance(where, str) or (
            ak._util.py27 and isinstance(where, ak._util.unicode)
        ):
            return self.replace_partitions([x[where] for x in self.partitions])

        elif isinstance(where, tuple) and len(where) == 0:
            return self

        elif (
            isinstance(where, Iterable)
            and len(where) > 0
            and all(
                (
                    isinstance(x, str)
                    or (ak._util.py27 and isinstance(x, ak._util.unicode))
                )
                for x in where
            )
        ):
            return self.replace_partitions([x[where] for x in self.partitions])

        else:
            if not isinstance(where, tuple):
                where = (where,)
            head, tail = where[0], where[1:]

            if not isinstance(head, (bool, np.bool_)) and isinstance(
                head, (numbers.Integral, np.integer)
            ):
                if head < 0:
                    head += len(self)
                if not 0 <= head < len(self):
                    raise ValueError(
                        "{0} index out of range".format(type(self).__name__)
                        + ak._util.exception_suffix(__file__)
                    )
                partitionid, index = self._ext.partitionid_index_at(head)
                return self.partition(partitionid)[(index,) + tail]

            elif isinstance(head, slice):
                partitions = self._ext.getitem_range(
                    head.start, head.stop, head.step
                ).partitions
                return IrregularlyPartitionedArray(
                    [x[(slice(None),) + tail] for x in partitions]
                )

            elif head is Ellipsis:
                return IrregularlyPartitionedArray(
                    [x[(head,) + tail] for x in self.partitions]
                )

            elif isinstance(head, str) or (
                ak._util.py27 and isinstance(head, ak._util.unicode)
            ):
                y = IrregularlyPartitionedArray([x[head] for x in self.partitions])
                if len(tail) == 0:
                    return y
                else:
                    return y[tail]

            elif (
                isinstance(head, Iterable)
                and len(head) > 0
                and all(
                    (
                        isinstance(x, str)
                        or (ak._util.py27 and isinstance(x, ak._util.unicode))
                    )
                    for x in head
                )
            ):
                y = IrregularlyPartitionedArray(
                    [x[list(head)] for x in self.partitions]
                )
                if len(tail) == 0:
                    return y
                else:
                    return y[tail]

            elif head is np.newaxis:
                return self.toContent()[(head,) + tail]

            else:
                layout = ak.operations.convert.to_layout(
                    head,
                    allow_record=False,
                    allow_other=False,
                    numpytype=(np.integer, np.bool_),
                )

                t = ak.operations.describe.type(layout)
                int_types = [
                    ak.types.PrimitiveType("int8"),
                    ak.types.PrimitiveType("uint8"),
                    ak.types.PrimitiveType("int16"),
                    ak.types.PrimitiveType("uint16"),
                    ak.types.PrimitiveType("int32"),
                    ak.types.PrimitiveType("uint32"),
                    ak.types.PrimitiveType("int64"),
                    ak.types.PrimitiveType("uint64"),
                ]
                opt_types = [ak.types.OptionType(p) for p in int_types]

                if (
                    isinstance(t, ak.types.OptionType)
                    or t in int_types + opt_types
                    or (
                        isinstance(t, ak.types.UnionType)
                        and all(ti in int_types for ti in t.types)
                    )
                    or (
                        isinstance(t, ak.types.OptionType)
                        and isinstance(t.type, ak.types.UnionType)
                        and all(ti in int_types for ti in t.type.types)
                    )
                ):
                    if isinstance(layout, PartitionedArray):
                        layout = layout.toContent()
                    return self.toContent()[(layout,) + tail]

                else:
                    stops = self.stops
                    if isinstance(layout, PartitionedArray):
                        layout = layout.repartition(stops)
                    else:
                        layout = IrregularlyPartitionedArray.toPartitioned(
                            layout, stops
                        )
                    inparts = self.partitions
                    headparts = layout.partitions
                    outparts = []
                    outoffsets = [0]
                    for i in range(len(inparts)):
                        outparts.append(inparts[i][(headparts[i],) + tail])
                        outoffsets.append(outoffsets[-1] + len(outparts[-1]))
                    return IrregularlyPartitionedArray(outparts, outoffsets[1:])

    def __iter__(self):
        for partition in self.partitions:
            for x in partition:
                yield x

    @property
    def kernels(self):
        out = None
        for x in self.partitions:
            if out is None:
                out = x.kernels
            elif out != x.kernels:
                return "mixed"
        if out is None:
            return "cpu"
        else:
            return out

    @property
    def caches(self):
        return list(ak._util.find_caches(self))

    def tojson(self, *args, **kwargs):
        return self._ext.tojson(*args, **kwargs)

    @property
    def nbytes(self):
        return sum(x.nbytes for x in self.partitions)

    def deep_copy(self, *args, **kwargs):
        out = type(self).__new__(type(self))
        out.__dict__.update(self.__dict__)
        out._partitions = [x.deep_copy(*args, **kwargs) for x in out.partitions]
        return out

    @property
    def identity(self):
        raise NotImplementedError(ak._util.exception_suffix(__file__))

    @property
    def numfields(self):
        return first(self).numfields

    def fieldindex(self, *args, **kwargs):
        return first(self).fieldindex(*args, **kwargs)

    def key(self, *args, **kwargs):
        return first(self).key(*args, **kwargs)

    def haskey(self, *args, **kwargs):
        return first(self).haskey(*args, **kwargs)

    def keys(self, *args, **kwargs):
        return first(self).keys(*args, **kwargs)

    @property
    def purelist_isregular(self):
        return first(self).purelist_isregular

    @property
    def purelist_depth(self):
        return first(self).purelist_depth

    @property
    def branch_depth(self):
        return first(self).branch_depth

    @property
    def minmax_depth(self):
        return first(self).minmax_depth

    def getitem_nothing(self, *args, **kwargs):
        return first(self).getitem_nothing(*args, **kwargs)

    def getitem_at_nowrap(self, where):
        assert not isinstance(where, (bool, np.bool_)) and isinstance(
            where, (numbers.Integral, np.integer)
        )
        assert where >= 0
        return self[where]

    def getitem_range_nowrap(self, start, stop):
        return self[start:stop]

    @property
    def _persistent_shared_ptr(self):
        raise NotImplementedError(ak._util.exception_suffix(__file__))

    def validityerror(self, *args, **kwargs):
        t = None
        for x in self.partitions:
            if t is None:
                t = x.type({})
            elif t != x.type({}):
                return "inconsistent types in PartitionedArray"
            out = x.validityerror(*args, **kwargs)
            if out is not None:
                return out
        return None

    def fillna(self, what):
        return self.replace_partitions([x.fillna(what) for x in self.partitions])

    def num(self, axis):
        if first(self).axis_wrap_if_negative(axis) == 0:
            prepared = [x.num(axis) for x in self.partitions]
            if any(isinstance(x, ak.layout.Record) for x in prepared):
                names = None
                counts = None
                for x in prepared:
                    assert isinstance(x, ak.layout.Record)
                    n = ak.operations.describe.fields(x)
                    c = ak.operations.structure.unzip(x)
                    if names is None:
                        names, counts = n, list(c)
                    else:
                        assert n == names
                        for i, ci in enumerate(c):
                            counts[i] += ci
                arraycounts = [ak.layout.NumpyArray(numpy.array([x])) for x in counts]
                return ak.layout.Record(ak.layout.RecordArray(arraycounts, names), 0)
            else:
                return sum(prepared)
        else:
            return self.replace_partitions([x.num(axis) for x in self.partitions])

    def flatten(self, *args, **kwargs):
        return apply(lambda x: x.flatten(*args, **kwargs), self)

    def offsets_and_flatten(self, axis):
        return self.toContent().offsets_and_flatten(axis)

    def rpad(self, length, axis):
        if first(self).axis_wrap_if_negative(axis) == 0:
            return self.toContent().rpad(length, axis)
        else:
            return self.replace_partitions(
                [x.rpad(length, axis) for x in self.partitions]
            )

    def rpad_and_clip(self, length, axis):
        if first(self).axis_wrap_if_negative(axis) == 0:
            return self.toContent().rpad_and_clip(length, axis)
        else:
            return self.replace_partitions(
                [x.rpad_and_clip(length, axis) for x in self.partitions]
            )

    def mergeable(self, other):
        if isinstance(other, PartitionedArray):
            other = other.toContent()
        return self.toContent().mergeable(other)

    def merge(self, other):
        if isinstance(other, PartitionedArray):
            other = other.toContent()
        return self.toContent().merge(other)

    def mergemany(self, others):
        others = [
            x.toContent() if isinstance(x, PartitionedArray) else x for x in others
        ]
        return self.toContent().mergemany(others)

    def axis_wrap_if_negative(self, axis):
        out = None
        for partition in self.partitions:
            if out is None:
                out = partition.axis_wrap_if_negative(axis)
            elif out != partition.axis_wrap_if_negative(axis):
                raise ValueError(
                    "partitions have inconsistent depths"
                    + ak._util.exception_suffix(__file__)
                )
        return out

    def reduce(self, name, axis, mask, keepdims, initial=None):
        branch, depth = first(self).branch_depth
        negaxis = -axis
        if not branch and negaxis <= 0:
            negaxis += depth
        if not branch and negaxis == depth:
            if initial is None:
                return getattr(self.toContent(), name)(axis, mask, keepdims)
            else:
                return getattr(self.toContent(), name)(axis, mask, keepdims, initial)
        else:
            return self.replace_partitions(
                [getattr(x, name)(axis, mask, keepdims) for x in self.partitions]
            )

    def count(self, axis, mask, keepdims):
        return self.reduce("count", axis, mask, keepdims)

    def count_nonzero(self, axis, mask, keepdims):
        return self.reduce("count_nonzero", axis, mask, keepdims)

    def sum(self, axis, mask, keepdims):
        return self.reduce("sum", axis, mask, keepdims)

    def prod(self, axis, mask, keepdims):
        return self.reduce("prod", axis, mask, keepdims)

    def any(self, axis, mask, keepdims):
        return self.reduce("any", axis, mask, keepdims)

    def all(self, axis, mask, keepdims):
        return self.reduce("all", axis, mask, keepdims)

    def min(self, axis, mask, keepdims, initial):
        return self.reduce("min", axis, mask, keepdims, initial)

    def max(self, axis, mask, keepdims, initial):
        return self.reduce("max", axis, mask, keepdims, initial)

    def argmin(self, axis, mask, keepdims):
        return self.reduce("argmin", axis, mask, keepdims)

    def argmax(self, axis, mask, keepdims):
        return self.reduce("argmax", axis, mask, keepdims)

    def localindex(self, axis):
        if first(self).axis_wrap_if_negative(axis) == 0:
            start = 0
            output = []
            for x in self.partitions:
                output.append(
                    ak.layout.NumpyArray(
                        ak.nplike.of(x).arange(start, start + len(x), dtype=np.int64)
                    )
                )
                start += len(x)
            return self.replace_partitions(output)

        else:
            return self.replace_partitions(
                [x.localindex(axis) for x in self.partitions]
            )

    def combinations(self, n, replacement, keys, parameters, axis):
        if first(self).axis_wrap_if_negative(axis) == 0:
            return self.toContent().combinations(n, replacement, keys, parameters, axis)
        else:
            return self.replace_partitions(
                [
                    x.combinations(n, replacement, keys, parameters, axis)
                    for x in self.partitions
                ]
            )

    def sort(self, axis, ascending, stable):
        if first(self).axis_wrap_if_negative(axis) == 0:
            return self.toContent().sort(axis, ascending, stable)
        else:
            return self.replace_partitions(
                [x.sort(axis, ascending, stable) for x in self.partitions]
            )

    def argsort(self, axis, ascending, stable):
        if first(self).axis_wrap_if_negative(axis) == 0:
            return self.toContent().argsort(axis, ascending, stable)
        else:
            return self.replace_partitions(
                [x.argsort(axis, ascending, stable) for x in self.partitions]
            )

    def numbers_to_type(self, dtype_string):
        return self.replace_partitions(
            [x.numbers_to_type(dtype_string) for x in self.partitions]
        )

    def is_unique(self):
        return all([x.is_unique() for x in self.partitions])

    def copy_to(self, ptr_lib):
        return self.from_ext(self._ext.copy_to(ptr_lib))

    def carry(self, carry, allow_lazy):
        return self.toContent().carry(carry, allow_lazy)


class IrregularlyPartitionedArray(PartitionedArray):
    @classmethod
    def toPartitioned(cls, layout, stops):
        partitions = []
        start = 0
        for stop in stops:
            partitions.append(layout[start:stop])
            start = stop
        return IrregularlyPartitionedArray(partitions, stops)

    def __init__(self, partitions, stops=None):
        if stops is None:
            nextpartitions = [x for x in partitions if len(x) != 0]
            if len(nextpartitions) == 0:
                nextpartitions = partitions[:1]
            self._ext = ak._ext.IrregularlyPartitionedArray(nextpartitions)
        else:
            nextpartitions = []
            nextstops = []
            start = 0
            for partition, stop in zip(partitions, stops):
                if stop - start != 0:
                    nextpartitions.append(partition)
                    nextstops.append(stop)
                start = stop
            if len(nextpartitions) == 0:
                nextpartitions = partitions[:1]
                nextstops = stops[:1]
            self._ext = ak._ext.IrregularlyPartitionedArray(nextpartitions, nextstops)

    @property
    def stops(self):
        return self._ext.stops

    def validityerror(self):
        total_length = 0
        for x, stop in zip(self.partitions, self.stops):
            total_length += len(x)
            if total_length != stop:
                return (
                    "IrregularlyPartitionedArray stops do not match "
                    "partition lengths"
                )
            out = x.validityerror()
            if out is not None:
                return out
        return None

    def replace_partitions(self, partitions):
        return IrregularlyPartitionedArray(partitions, self.stops)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pandas = pytest.importorskip("pandas")


def test():
    simple = ak.Array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
    assert ak.to_pandas(simple)["values"].values.tolist() == [
        0.0,
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]

    index = ak.layout.Index64(np.array([3, 3, 1, 5], dtype=np.int64))
    indexed = ak.Array(ak.layout.IndexedArray64(index, simple.layout))
    assert indexed.tolist() == [3.3, 3.3, 1.1, 5.5]

    assert ak.to_pandas(indexed)["values"].values.tolist() == [3.3, 3.3, 1.1, 5.5]

    tuples = ak.Array(ak.layout.RecordArray([simple.layout, simple.layout]))
    assert ak.to_pandas(tuples)["1"].values.tolist() == [0.0, 1.1, 2.2, 3.3, 4.4, 5.5]

    offsets = ak.layout.Index64(np.array([0, 1, 1, 3, 4], dtype=np.int64))
    nested = ak.Array(ak.layout.ListOffsetArray64(offsets, indexed.layout))
    assert ak.to_pandas(nested)["values"].values.tolist() == [3.3, 3.3, 1.1, 5.5]

    offsets2 = ak.layout.Index64(np.array([0, 3, 3, 4, 6], dtype=np.int64))
    nested2 = ak.Array(ak.layout.ListOffsetArray64(offsets2, tuples.layout))

    assert ak.to_pandas(nested2)["1"].values.tolist() == [0.0, 1.1, 2.2, 3.3, 4.4, 5.5]

    recrec = ak.Array([{"x": {"y": 1}}, {"x": {"y": 2}}, {"x": {"y": 3}}])
    assert ak.to_pandas(recrec)["x", "y"].values.tolist() == [1, 2, 3]

    recrec2 = ak.Array(
        [
            {"x": {"a": 1, "b": 2}, "y": {"c": 3, "d": 4}},
            {"x": {"a": 10, "b": 20}, "y": {"c": 30, "d": 40}},
        ]
    )
    assert ak.to_pandas(recrec2)["y", "c"].values.tolist() == [3, 30]

    recrec3 = ak.Array(
        [{"x": 1, "y": {"c": 3, "d": 4}}, {"x": 10, "y": {"c": 30, "d": 40}}]
    )
    assert ak.to_pandas(recrec3)["y", "c"].values.tolist() == [3, 30]

    tuptup = ak.Array([(1.0, (1.1, 1.2)), (2.0, (2.1, 2.2)), (3.0, (3.1, 3.2))])
    assert ak.to_pandas(tuptup)["1", "0"].values.tolist() == [1.1, 2.1, 3.1]

    recrec4 = ak.Array(
        [[{"x": 1, "y": {"c": 3, "d": 4}}], [{"x": 10, "y": {"c": 30, "d": 40}}]]
    )
    assert ak.to_pandas(recrec4)["y", "c"].values.tolist() == [3, 30]


def test_broken():
    ex = ak.Array([[1, 2, 3], [], [4, 5]])
    p4 = ak.zip({"x": ex})
    p4c = ak.cartesian({"a": p4, "b": p4})
    df = ak.to_pandas(p4c)
    assert df["a", "x"].values.tolist() == [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 5, 5]
    assert df["b", "x"].values.tolist() == [1, 2, 3, 1, 2, 3, 1, 2, 3, 4, 5, 4, 5]


def test_union_to_record():
    recordarray1 = ak.Array([{"x": 1, "y": 1.1}, {"x": 3, "y": 3.3}]).layout
    recordarray2 = ak.Array([{"y": 2.2, "z": 999}]).layout
    tags = ak.layout.Index8(np.array([0, 1, 0], dtype=np.int8))
    index = ak.layout.Index64(np.array([0, 0, 1], dtype=np.int64))
    unionarray = ak.layout.UnionArray8_64(tags, index, [recordarray1, recordarray2])
    assert ak.to_list(unionarray) == [
        {"x": 1, "y": 1.1},
        {"y": 2.2, "z": 999},
        {"x": 3, "y": 3.3},
    ]

    converted = ak._util.union_to_record(unionarray, "values")
    assert isinstance(converted, ak.layout.RecordArray)
    assert ak.to_list(converted) == [
        {"x": 1, "y": 1.1, "z": None},
        {"x": None, "y": 2.2, "z": 999},
        {"x": 3, "y": 3.3, "z": None},
    ]

    otherarray = ak.Array(["one", "two"]).layout
    tags2 = ak.layout.Index8(np.array([0, 2, 1, 2, 0], dtype=np.int8))
    index2 = ak.layout.Index64(np.array([0, 0, 0, 1, 1], dtype=np.int64))
    unionarray2 = ak.layout.UnionArray8_64(
        tags2, index2, [recordarray1, recordarray2, otherarray]
    )
    assert ak.to_list(unionarray2) == [
        {"x": 1, "y": 1.1},
        "one",
        {"y": 2.2, "z": 999},
        "two",
        {"x": 3, "y": 3.3},
    ]

    converted2 = ak._util.union_to_record(unionarray2, "values")
    assert isinstance(converted2, ak.layout.RecordArray)
    assert ak.to_list(converted2) == [
        {"x": 1, "y": 1.1, "z": None, "values": None},
        {"x": None, "y": None, "z": None, "values": "one"},
        {"x": None, "y": 2.2, "z": 999, "values": None},
        {"x": None, "y": None, "z": None, "values": "two"},
        {"x": 3, "y": 3.3, "z": None, "values": None},
    ]

    df_unionarray = ak.to_pandas(unionarray)
    np.testing.assert_array_equal(df_unionarray["x"].values, np.array([1, np.nan, 3]))
    np.testing.assert_array_equal(df_unionarray["y"].values, np.array([1.1, 2.2, 3.3]))
    np.testing.assert_array_equal(
        df_unionarray["z"].values, np.array([np.nan, 999, np.nan])
    )

    df_unionarray2 = ak.to_pandas(unionarray2)
    np.testing.assert_array_equal(
        df_unionarray2["x"].values, [1, np.nan, np.nan, np.nan, 3]
    )
    np.testing.assert_array_equal(
        df_unionarray2["y"].values, [1.1, np.nan, 2.2, np.nan, 3.3]
    )
    np.testing.assert_array_equal(
        df_unionarray2["z"].values, [np.nan, np.nan, 999, np.nan, np.nan]
    )
    np.testing.assert_array_equal(
        df_unionarray2["values"].values, ["nan", "one", "nan", "two", "nan"]
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os
import datetime
import time

import yaml

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))


def reproducible_datetime():

    build_date = datetime.datetime.utcfromtimestamp(
        int(os.environ.get("SOURCE_DATE_EPOCH", time.time()))
    )
    return build_date.isoformat().replace("T", " AT ")[:22]


def type_to_ctype(typename):
    is_const = False
    if "Const[" in typename:
        is_const = True
        typename = typename[len("Const[") : -1]
    count = 0
    while "List[" in typename:
        count += 1
        typename = typename[len("List[") : -1]
    typename = typename + "*" * count
    if is_const:
        typename = "const " + typename
    return typename


def include_kernels_h(specification):
    print("Generating include/awkward/kernels.h...")

    with open(
        os.path.join(CURRENT_DIR, "..", "include", "awkward", "kernels.h"), "w"
    ) as header:
        header.write(
            """// AUTO GENERATED ON {0}
// DO NOT EDIT BY HAND!
//
// To regenerate file, run
//
//     python dev/generate-kernel-signatures.py
//
// (It is usually run as part of pip install . or localbuild.py.)

#ifndef AWKWARD_KERNELS_H_
#define AWKWARD_KERNELS_H_

#include "awkward/common.h"

extern "C" {{

""".format(
                reproducible_datetime()
            )
        )
        for spec in specification["kernels"]:
            for childfunc in spec["specializations"]:
                header.write(" " * 2 + "EXPORT_SYMBOL ERROR\n")
                header.write(" " * 2 + childfunc["name"] + "(\n")
                for i, arg in enumerate(childfunc["args"]):
                    header.write(
                        " " * 4 + type_to_ctype(arg["type"]) + " " + arg["name"]
                    )
                    if i == (len(childfunc["args"]) - 1):
                        header.write(");\n")
                    else:
                        header.write(",\n")
            header.write("\n")
        header.write(
            """}

#endif // AWKWARD_KERNELS_H_
"""
        )

    print("Done with  include/awkward/kernels.h.")


type_to_dtype = {
    "bool": "bool_",
    "int8": "int8",
    "uint8": "uint8",
    "int16": "int16",
    "uint16": "uint16",
    "int32": "int32",
    "uint32": "uint32",
    "int64": "int64",
    "uint64": "uint64",
    "float": "float32",
    "double": "float64",
}


def type_to_pytype(typename, special):
    if "Const[" in typename:
        typename = typename[len("Const[") : -1]
    count = 0
    while "List[" in typename:
        count += 1
        typename = typename[len("List[") : -1]
    if typename.endswith("_t"):
        typename = typename[:-2]
    if count != 0:
        special.append(type_to_dtype[typename])
    return ("POINTER(" * count) + ("c_" + typename) + (")" * count)


def kernel_signatures_py(specification):
    print("Generating src/awkward/_kernel_signatures.py...")

    with open(
        os.path.join(CURRENT_DIR, "..", "src", "awkward", "_kernel_signatures.py"),
        "w",
    ) as file:
        file.write(
            """# AUTO GENERATED ON {0}
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-kernel-signatures.py
#
# (It is usually run as part of pip install . or localbuild.py.)

# fmt: off

from ctypes import (
    POINTER,
    Structure,
    c_bool,
    c_int8,
    c_uint8,
    c_int16,
    c_uint16,
    c_int32,
    c_uint32,
    c_int64,
    c_uint64,
    c_float,
    c_double,
    c_char_p,
)

import numpy as np

from numpy import (
    bool_,
    int8,
    uint8,
    int16,
    uint16,
    int32,
    uint32,
    int64,
    uint64,
    float32,
    float64,
)

class ERROR(Structure):
    _fields_ = [
        ("str", c_char_p),
        ("filename", c_char_p),
        ("id", c_int64),
        ("attempt", c_int64),
        ("pass_through", c_bool),
    ]


def by_signature(lib):
    out = {{}}
""".format(
                reproducible_datetime()
            )
        )

        for spec in specification["kernels"]:
            for childfunc in spec["specializations"]:
                special = [repr(spec["name"])]
                arglist = [
                    type_to_pytype(x["type"], special) for x in childfunc["args"]
                ]
                dirlist = [repr(x["dir"]) for x in childfunc["args"]]
                file.write(
                    """
    f = lib.{0}
    f.argtypes = [{1}]
    f.restype = ERROR
    f.dir = [{2}]
    out[{3}] = f
""".format(
                        childfunc["name"],
                        ", ".join(arglist),
                        ", ".join(dirlist),
                        ", ".join(special),
                    )
                )

        file.write(
            """
    return out
"""
        )

    print("Done with  src/awkward/_kernel_signatures.py...")


if __name__ == "__main__":
    with open(os.path.join(CURRENT_DIR, "..", "kernel-specification.yml")) as specfile:
        specification = yaml.safe_load(specfile)
        include_kernels_h(specification)
        kernel_signatures_py(specification)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_regular():
    np_data = np.array(
        [[1, 3, 5, 4, 2], [3, 7, 8, 2, 4], [2, 3, 1, 7, 7], [5, 1, 9, 10, 2]]
    )
    ak_data = ak.from_numpy(np_data)

    assert ak.ptp(ak_data, axis=1).tolist() == np.ptp(np_data, axis=1).tolist()
    assert ak.ptp(ak_data, axis=0).tolist() == np.ptp(np_data, axis=0).tolist()
    assert ak.ptp(ak_data) == np.ptp(np_data)


def test_jagged():
    data = ak.Array(
        [
            [1, 3, 5, 4, 2],
            [],
            [2, 3, 1],
            [5],
        ]
    )
    assert ak.ptp(data, axis=1, mask_identity=False).tolist() == [4, 0, 2, 0]
    assert ak.ptp(data, axis=1).tolist() == [4, None, 2, 0]
    assert ak.ptp(data, axis=0).tolist() == [4, 0, 4, 0, 0]
    assert ak.ptp(data) == 4



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak
from awkward._v2.types.type import Type
from awkward._v2.forms.form import _parameters_equal


class UnionType(Type):
    def __init__(self, contents, parameters=None, typestr=None):
        if not isinstance(contents, Iterable):
            raise TypeError(
                "{0} 'contents' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if not isinstance(contents, list):
            contents = list(contents)
        for content in contents:
            if not isinstance(content, Type):
                raise TypeError(
                    "{0} all 'contents' must be Type subclasses, not {1}".format(
                        type(self).__name__, repr(content)
                    )
                )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{0} 'parameters' must be of type dict or None, not {1}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if typestr is not None and not ak._util.isstr(typestr):
            raise TypeError(
                "{0} 'typestr' must be of type string or None, not {1}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._contents = contents
        self._parameters = parameters
        self._typestr = typestr

    @property
    def contents(self):
        return self._contents

    def __str__(self):
        if self._typestr is not None:
            out = self._typestr

        else:
            children = [str(x) for x in self._contents]
            params = self._str_parameters()

            if params is None:
                out = "union[{0}]".format(", ".join(children))
            else:
                out = "union[{0}, {1}]".format(", ".join(children), params)

        return self._str_categorical_begin() + out + self._str_categorical_end()

    def __repr__(self):
        args = [repr(self._contents)] + self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def __eq__(self, other):
        if isinstance(other, UnionType):
            return (
                self._typestr == other._typestr
                and _parameters_equal(self._parameters, other._parameters)
                and self._contents == other._contents
            )
        else:
            return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array([1j])
    assert array.tolist() == [1j]

    array1 = ak.concatenate(([3j], [3j]))
    assert array1.tolist() == [3j, 3j]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
import awkward._v2.types.type
from awkward._v2.forms.form import _parameters_equal


class ArrayType(object):
    def __init__(self, content, length):
        if not isinstance(content, awkward._v2.types.type.Type):
            raise TypeError(
                "{0} all 'contents' must be Type subclasses, not {1}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not ak._util.isint(length) or length < 0:
            raise ValueError(
                "{0} 'length' must be of a positive integer, not {1}".format(
                    type(self).__name__, repr(length)
                )
            )
        self._content = content
        self._length = length

    @property
    def content(self):
        return self._content

    @property
    def length(self):
        return self._length

    def __str__(self):
        return str(self._length) + " * " + str(self._content)

    def __repr__(self):
        args = [repr(self._content), repr(self._length)]
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def __eq__(self, other):
        if isinstance(other, ArrayType):
            return (
                self._length == other._length
                and _parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.to_categorical(ak.Array([321, 1.1, 123, 1.1, 999, 1.1, 2.0]))
    assert ak.to_list(array * 10) == [3210, 11, 1230, 11, 9990, 11, 20]

    array = ak.Array(["HAL"])
    with pytest.raises(ValueError):
        array + 1



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


class Cache(MutableMapping):
    def __init__(self):
        self.data = {}

    def __getitem__(self, where):
        return self.data[where]

    def __setitem__(self, where, what):
        self.data[where] = what

    def __delitem__(self, where):
        del self.data[where]

    def __iter__(self):
        for x in self.data:
            yield x

    def __len__(self):
        return len(self.data)


def make_arrays():
    cache = Cache()
    x = ak.layout.VirtualArray(
        ak.layout.ArrayGenerator(
            lambda: ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3]))
        ),
        ak.layout.ArrayCache(cache),
    )
    y = ak.layout.VirtualArray(
        ak.layout.ArrayGenerator(
            lambda: ak.layout.NumpyArray(np.array([100, 200, 300]))
        ),
        ak.layout.ArrayCache(cache),
    )
    inner = ak.layout.RecordArray({"x": x, "y": y})
    part = ak.partition.IrregularlyPartitionedArray([inner], [3])
    return ak.Array(part)


def test():
    z = make_arrays()
    repr(z)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_empty_slice():
    # muon = ak.Array([[{"pt": 1.0}], []], with_name="muon")
    electron = ak.Array([[], [{"pt": 1.0}]], with_name="electron")

    electron = electron[electron.pt > 5]

    id = ak.argsort(electron, axis=1)

    assert ak.to_list(electron[id]) == [[], []]


# from coffea import nanoevents
#
# def test_empty_unions():
#     nanoevents.NanoAODSchema.mixins["FatJetLS"] = "PtEtaPhiMLorentzVector"
#
#     x = nanoevents.NanoEventsFactory.from_root(
#         './nano106Xv8_on_mini106X_2017_mc_NANO_py_NANO_7.root',
#         entry_start=100560,
#         entry_stop=102266
#     )
#
#     events = x.events()
#
#     muons = events.Muon
#     electrons = events.Electron
#     jets = events.Jet
#     fatjets = events.FatJet
#     subjets = events.SubJet
#     fatjetsLS = events.FatJetLS
#     met = events.MET
#
#     goodmuon = (
#         (muons.mediumId)
#         & (muons.miniPFRelIso_all <= 0.2)
#         & (muons.pt >= 27)
#         & (abs(muons.eta) <= 2.4)
#         & (abs(muons.dz) < 0.1)
#         & (abs(muons.dxy) < 0.05)
#         & (muons.sip3d < 4)
#     )
#     good_muons = muons[goodmuon]
#
#     # electrons
#     goodelectron = (
#         (electrons.mvaFall17V2noIso_WP90)
#         & (electrons.pt >= 30)
#         & (abs(electrons.eta) <= 1.479)
#         & (abs(electrons.dz) < 0.1)
#         & (abs(electrons.dxy) < 0.05)
#         & (electrons.sip3d < 4)
#     )
#     good_electrons = electrons[goodelectron]
#
#     leptons = ak.concatenate([muons, electrons], axis=1)
#     good_leptons = ak.concatenate([good_muons, good_electrons], axis=1)
#
#     idx1 = ak.argsort(leptons.pt, axis=1)
#     idx2 = ak.argsort(good_leptons.pt, axis=1)
#
#     leptons[idx1]
#
#     good_leptons[idx2]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_empty_listarray():
    a = ak.Array(
        ak.layout.ListArray64(
            ak.layout.Index64(np.array([], dtype=np.int64)),
            ak.layout.Index64(np.array([], dtype=np.int64)),
            ak.layout.NumpyArray(np.array([])),
        )
    )
    assert ak.to_list(a * 3) == []

    starts = ak.layout.Index64(np.array([], dtype=np.int64))
    stops = ak.layout.Index64(np.array([3, 3, 5], dtype=np.int64))
    content = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    array = ak.Array(ak.layout.ListArray64(starts, stops, content))
    array + array



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    layout = ak.layout.ListOffsetArray64(
        ak.layout.Index64(np.array([0, 1], dtype=np.int64)),
        ak.layout.IndexedArray64(
            ak.layout.Index64(np.array([0, 1, 2, 3], dtype=np.int64)),
            ak.layout.RegularArray(
                ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4])), 4
            ),
        ),
    )
    array = ak.Array(layout)

    assert str(ak.min(array, axis=-1, mask_identity=False).type) == "1 * var * float64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_record():
    assert ak.to_list(ak.Record({"x": 1, "y": 2.2})) == {"x": 1, "y": 2.2}


def test_fromiter():
    array = ak.Array([np.array([1, 2, 3]), np.array([4, 5, 6, 7])])

    assert str(ak.type(array)) == "2 * var * int64"
    assert ak.to_list(array) == [[1, 2, 3], [4, 5, 6, 7]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2.index import Index
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.indexedform import IndexedForm

np = ak.nplike.NumpyMetadata.instance()


class IndexedArray(Content):
    def __init__(self, index, content, identifier=None, parameters=None):
        if not (
            isinstance(index, Index)
            and index.dtype
            in (
                np.dtype(np.int32),
                np.dtype(np.uint32),
                np.dtype(np.int64),
            )
        ):
            raise TypeError(
                "{0} 'index' must be an Index with dtype in (int32, uint32, int64), "
                "not {1}".format(type(self).__name__, repr(index))
            )
        if not isinstance(content, Content):
            raise TypeError(
                "{0} 'content' must be a Content subtype, not {1}".format(
                    type(self).__name__, repr(content)
                )
            )

        self._index = index
        self._content = content
        self._init(identifier, parameters)

    @property
    def index(self):
        return self._index

    @property
    def content(self):
        return self._content

    @property
    def nplike(self):
        return self._index.nplike

    @property
    def nonvirtual_nplike(self):
        return self._index.nplike

    Form = IndexedForm

    @property
    def form(self):
        return self.Form(
            self._index.form,
            self._content.form,
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return IndexedArray(
            ak._v2.index.Index(self._index.to(tt)),
            self._content.typetracer,
            self._typetracer_identifier(),
            self._parameters,
        )

    def __len__(self):
        return len(self._index)

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<IndexedArray len="]
        out.append(repr(str(len(self))))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._index._repr(indent + "    ", "<index>", "</index>\n"))
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</IndexedArray>")
        out.append(post)
        return "".join(out)

    def toIndexedOptionArray64(self):
        return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
            self._index, self._content, self._identifier, self._parameters
        )

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if where < 0:
            where += len(self)
        if not (0 <= where < len(self)) and self.nplike.known_shape:
            raise NestedIndexError(self, where)
        return self._content._getitem_at(self._index[where])

    def _getitem_range(self, where):
        start, stop, step = where.indices(len(self))
        assert step == 1
        return IndexedArray(
            self._index[start:stop],
            self._content,
            self._range_identifier(start, stop),
            self._parameters,
        )

    def _getitem_field(self, where, only_fields=()):
        return IndexedArray(
            self._index,
            self._content._getitem_field(where, only_fields),
            self._field_identifier(where),
            None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return IndexedArray(
            self._index,
            self._content._getitem_fields(where, only_fields),
            self._fields_identifier(where),
            None,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        try:
            nextindex = self._index[carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))

        return IndexedArray(
            nextindex,
            self._content,
            self._carry_identifier(carry, exception),
            self._parameters,
        )

    def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
        nplike = self.nplike
        if len(slicestarts) != len(self) and nplike.known_shape:
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(slicestarts, slicestops, slicecontent),
                "cannot fit jagged slice with length {0} into {1} of size {2}".format(
                    len(slicestarts), type(self).__name__, len(self)
                ),
            )

        nextcarry = ak._v2.index.Index64.zeros(len(self), nplike)
        self._handle_error(
            nplike[
                "awkward_IndexedArray_getitem_nextcarry",
                nextcarry.dtype.type,
                self._index.dtype.type,
            ](
                nextcarry.to(nplike),
                self._index.to(nplike),
                len(self._index),
                len(self._content),
            )
        )
        # an eager carry (allow_lazy = false) to avoid infinite loop (unproven)
        next = self._content._carry(nextcarry, False, NestedIndexError)
        return next._getitem_next_jagged(slicestarts, slicestops, slicecontent, tail)

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        return self._getitem_next_jagged_generic(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(self, head, tail, advanced):
        nplike = self.nplike

        if head == ():
            return self

        elif isinstance(head, (int, slice, ak._v2.index.Index64)):
            nexthead, nexttail = ak._v2._slicing.headtail(tail)

            nextcarry = ak._v2.index.Index64.empty(len(self._index), nplike)
            self._handle_error(
                nplike[
                    "awkward_IndexedArray_getitem_nextcarry",
                    nextcarry.dtype.type,
                    self._index.dtype.type,
                ](
                    nextcarry.to(nplike),
                    self._index.to(nplike),
                    len(self._index),
                    len(self._content),
                )
            )

            next = self._content._carry(nextcarry, False, NestedIndexError)
            return next._getitem_next(head, tail, advanced)

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            raise NotImplementedError

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def project(self):
        nextcarry = ak._v2.index.Index64.empty(len(self), self.nplike)
        self._handle_error(
            self.nplike[
                "awkward_IndexedArray_getitem_nextcarry",
                nextcarry.dtype.type,
                self._index.dtype.type,
            ](
                nextcarry.to(self.nplike),
                self._index.to(self.nplike),
                len(self._index),
                len(self._content),
            )
        )
        return self._content._carry(nextcarry, False, NestedIndexError)

    def _localindex(self, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        else:
            return self.project()._localindex(posaxis, depth)

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if len(self._index) == 0:
            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))

        next = self._content._carry(self._index, False, NestedIndexError)
        return next._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _sort_next(
        self,
        negaxis,
        starts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        next = self._content._carry(self._index, False, NestedIndexError)
        return next._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            return self.project()._combinations(
                n, replacement, recordlookup, parameters, posaxis, depth
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        nplike = self.nplike
        branch, depth = self.branch_depth

        index_length = len(self._index)

        nextcarry = ak._v2.index.Index64.zeros(index_length, nplike)
        nextparents = ak._v2.index.Index64.zeros(index_length, nplike)
        outindex = ak._v2.index.Index64.zeros(index_length, nplike)
        self._handle_error(
            nplike[
                "awkward_IndexedArray_reduce_next_64",
                nextcarry.dtype.type,
                nextparents.dtype.type,
                outindex.dtype.type,
                self._index.dtype.type,
                parents.dtype.type,
            ](
                nextcarry.to(nplike),
                nextparents.to(nplike),
                outindex.to(nplike),
                self._index.to(nplike),
                parents.to(nplike),
                index_length,
            )
        )
        next = self._content._carry(nextcarry, False, NestedIndexError)
        nextshifts = None
        out = next._reduce_next(
            reducer,
            negaxis,
            starts,
            nextshifts,
            nextparents,
            outlength,
            mask,
            keepdims,
        )

        if not branch and negaxis == depth:
            return out
        else:
            if isinstance(out, ak._v2.contents.RegularArray):
                out = out.toListOffsetArray64(True)

            elif isinstance(out, ak._v2.contents.ListOffsetArray):
                if len(starts) > 0 and starts[0] != 0:
                    raise AssertionError(
                        "reduce_next with unbranching depth > negaxis expects a "
                        "ListOffsetArray64 whose offsets start at zero ({0})".format(
                            starts[0]
                        )
                    )

                outoffsets = ak._v2.index.Index64.empty(len(starts) + 1, nplike)
                self._handle_error(
                    nplike[
                        "awkward_IndexedArray_reduce_next_fix_offsets_64",
                        outoffsets.dtype.type,
                        starts.dtype.type,
                    ](
                        outoffsets.to(nplike),
                        starts.to(nplike),
                        len(starts),
                        len(self._index),
                    )
                )

                tmp = ak._v2.contents.IndexedArray(
                    outindex,
                    out._content,
                    None,
                    None,
                )

                return ak._v2.contents.ListOffsetArray(
                    outoffsets,
                    tmp,
                    None,
                    None,
                )

            else:
                raise AssertionError(
                    "reduce_next with unbranching depth > negaxis is only "
                    "expected to return RegularArray or ListOffsetArray64; "
                    "instead, it returned " + out.classname
                )

    def _validityerror(self, path):
        error = self.nplike["awkward_IndexedArray_validity", self.index.dtype.type](
            self.index.to(self.nplike), len(self.index), len(self.content), False
        )
        if error.str is not None:
            if error.filename is None:
                filename = ""
            else:
                filename = " (in compiled code: " + error.filename.decode(
                    errors="surrogateescape"
                ).lstrip("\n").lstrip("(")
            message = error.str.decode(errors="surrogateescape")
            return 'at {0} ("{1}"): {2} at i={3}{4}'.format(
                path, type(self), message, error.id, filename
            )

        elif isinstance(
            self.content,
            (
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return "{0} contains \"{1}\", the operation that made it might have forgotten to call 'simplify_optiontype()'"
        else:
            return self.content.validityerror(path + ".content")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_array_builder():
    builder = ak.ArrayBuilder()

    builder.real(1.1)
    builder.real(2.2)
    builder.real(3.3)

    assert ak.to_list(builder.snapshot()) == [1.1, 2.2, 3.3]


def test_array_bool_builder():
    builder = ak.ArrayBuilder()

    builder.boolean(True)
    builder.boolean(False)
    builder.boolean(True)

    assert ak.to_list(builder.snapshot()) == [True, False, True]


def test_array_builder_append():
    array = ak.Array(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], check_valid=True
    )

    builder = ak.ArrayBuilder()
    builder.append(array, 3)
    builder.append(array, 2)
    builder.append(array, 2)
    builder.append(array, 0)
    builder.append(array, 1)
    builder.append(array, -1)
    assert ak.to_list(builder.snapshot()) == [
        [5.5],
        [3.3, 4.4],
        [3.3, 4.4],
        [0.0, 1.1, 2.2],
        [],
        [6.6, 7.7, 8.8, 9.9],
    ]

    builder.extend(array)
    assert ak.to_list(builder.snapshot()) == [
        [5.5],
        [3.3, 4.4],
        [3.3, 4.4],
        [0.0, 1.1, 2.2],
        [],
        [6.6, 7.7, 8.8, 9.9],
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2.forms.form import (
    Form,
    _parameters_equal,
    _parameters_update,
    nonvirtual,
)


class IndexedForm(Form):
    def __init__(
        self,
        index,
        content=None,
        has_identifier=False,
        parameters=None,
        form_key=None,
    ):
        if not ak._util.isstr(index):
            raise TypeError(
                "{0} 'index' must be of type str, not {1}".format(
                    type(self).__name__, repr(index)
                )
            )
        if not isinstance(content, Form):
            raise TypeError(
                "{0} all 'contents' must be Form subclasses, not {1}".format(
                    type(self).__name__, repr(content)
                )
            )

        self._index = index
        self._content = content
        self._init(has_identifier, parameters, form_key)

    @property
    def index(self):
        return self._index

    @property
    def content(self):
        return self._content

    def __repr__(self):
        args = [repr(self._index), repr(self._content)] + self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def _tolist_part(self, verbose, toplevel):
        return self._tolist_extra(
            {
                "class": "IndexedArray",
                "index": self._index,
                "content": self._content._tolist_part(verbose, toplevel=False),
            },
            verbose,
        )

    def _type(self, typestrs):
        out = self._content._type(typestrs)

        if self._parameters is not None:
            if out._parameters is None:
                out._parameters = self._parameters
            else:
                out._parameters = dict(out._parameters)
                _parameters_update(out._parameters, self._parameters)

            if self._parameters.get("__array__") == "categorical":
                if out._parameters is self._parameters:
                    out._parameters = dict(out._parameters)
                del out._parameters["__array__"]
                out._parameters["__categorical__"] = True

        return out

    def __eq__(self, other):
        if isinstance(other, IndexedForm):
            return (
                self._has_identifier == other._has_identifier
                and self._form_key == other._form_key
                and self._index == other._index
                and _parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False

    def generated_compatibility(self, other):
        other = nonvirtual(other)

        if other is None:
            return True

        elif isinstance(other, IndexedForm):
            return (
                self._index == other._index
                and _parameters_equal(self._parameters, other._parameters)
                and self._content.generated_compatibility(other._content)
            )

        else:
            return False

    def _getitem_range(self):
        return IndexedForm(
            self._index,
            self._content,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def _getitem_field(self, where, only_fields=()):
        return IndexedForm(
            self._index,
            self._content._getitem_field(where, only_fields),
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return IndexedForm(
            self._index,
            self._content._getitem_fields(where, only_fields),
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _carry(self, allow_lazy):
        return IndexedForm(
            self._index,
            self._content,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def purelist_isregular(self):
        return self._content.purelist_isregular

    @property
    def purelist_depth(self):
        return self._content.purelist_depth

    @property
    def minmax_depth(self):
        return self._content.minmax_depth

    @property
    def branch_depth(self):
        return self._content.branch_depth

    @property
    def keys(self):
        return self._content.keys

    @property
    def dimension_optiontype(self):
        return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    a = ak.values_astype(ak.Array([1, None]), np.float32)

    assert ak.fill_none(a, np.float32(0)).tolist() == [1, 0]
    assert str(ak.fill_none(a, np.float32(0)).type) == "2 * float32"

    assert ak.fill_none(a, np.array(0, np.float32)).tolist() == [1, 0]
    assert str(ak.fill_none(a, np.array(0, np.float32)).type) == "2 * float32"

    assert ak.fill_none(a, np.array([0], np.float32)).tolist() == [1, [0]]
    assert (
        str(ak.fill_none(a, np.array([0], np.float32)).type)
        == "2 * union[float32, 1 * float32]"
    )

    assert ak.fill_none(a, np.array([[0]], np.float32)).tolist() == [1, [[0]]]
    assert (
        str(ak.fill_none(a, np.array([[0]], np.float32)).type)
        == "2 * union[float32, 1 * 1 * float32]"
    )

    assert ak.fill_none(a, 0).tolist() == [1, 0]
    assert str(ak.fill_none(a, 0).type) == "2 * float64"

    assert ak.fill_none(a, [0]).tolist() == [1, [0]]
    assert str(ak.fill_none(a, [0]).type) == "2 * union[float32, 1 * int64]"

    assert ak.fill_none(a, [[0]]).tolist() == [1, [[0]]]
    assert str(ak.fill_none(a, [[0]]).type) == "2 * union[float32, 1 * var * int64]"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    parameters = ak.from_numpy(np.arange(3 * 3).reshape(-1, 3), regulararray=True)
    array = ak.zip((parameters, parameters))
    assert array.layout.purelist_isregular

    assert ak.to_list(array) == [
        [(0, 0), (1, 1), (2, 2)],
        [(3, 3), (4, 4), (5, 5)],
        [(6, 6), (7, 7), (8, 8)],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


vector = pytest.importorskip("vector")


def awkward_isnt_installed_globally():
    import importlib_metadata

    try:
        importlib_metadata.version("awkward")
    except importlib_metadata.PackageNotFoundError:
        return True
    else:
        return False


@pytest.mark.skipif(
    awkward_isnt_installed_globally(),
    reason="Awkward Array isn't installed globally for Vector",
)
def test():
    point = ak.Record(
        {"x": 1, "y": 2, "z": 3},
        with_name="Vector3D",
        behavior=vector._backends.awkward_.behavior,
    )
    assert np.matmul(point, vector.obj(x=1, y=0, z=2)) == 7



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_fillna_empty_array():
    empty = ak.layout.EmptyArray()

    assert ak.to_list(empty) == []
    array = empty.rpad(5, 0)
    assert ak.to_list(array) == [None, None, None, None, None]
    value = ak.layout.NumpyArray(np.array([10]))
    assert ak.to_list(array.fillna(value)) == [10, 10, 10, 10, 10]


def test_fillna_numpy_array():
    content = ak.layout.NumpyArray(np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]]))
    array = content.rpad(3, 0)
    assert ak.to_list(array) == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], None]
    value = ak.layout.NumpyArray(np.array([0]))
    assert ak.to_list(array.fillna(value)) == [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], 0]

    array = content.rpad(5, 1)
    assert ak.to_list(array) == [
        [1.1, 2.2, 3.3, None, None],
        [4.4, 5.5, 6.6, None, None],
    ]
    value = ak.layout.NumpyArray(np.array([0]))
    assert ak.to_list(array.fillna(value)) == [
        [1.1, 2.2, 3.3, 0, 0],
        [4.4, 5.5, 6.6, 0, 0],
    ]


# def test_fillna_numpy_O_array():
#     pyobject_array = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, None, 5.5, 6.6, None]))
#     assert ak.to_list(optarray.fillna(0)) == [1.1, 2.2, 3.3, 0, 5.5, 6.6, 0]


def test_fillna_regular_array():
    content = ak.layout.NumpyArray(
        np.array(
            [
                2.1,
                8.4,
                7.4,
                1.6,
                2.2,
                3.4,
                6.2,
                5.4,
                1.5,
                3.9,
                3.8,
                3.0,
                8.5,
                6.9,
                4.3,
                3.6,
                6.7,
                1.8,
                3.2,
            ]
        )
    )
    index = ak.layout.Index64(
        np.array([13, 9, 13, 4, 8, 3, 15, -1, 16, 2, 8], dtype=np.int64)
    )
    indexedarray = ak.layout.IndexedOptionArray64(index, content)
    regarray = ak.layout.RegularArray(indexedarray, 3, zeros_length=0)

    assert ak.to_list(regarray) == [[6.9, 3.9, 6.9], [2.2, 1.5, 1.6], [3.6, None, 6.7]]
    value = ak.layout.NumpyArray(np.array([666]))
    assert ak.to_list(regarray.fillna(value)) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, 666, 6.7],
    ]


def test_fillna_listarray_array():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.layout.Index64(np.array([0, 3, 4, 5, 8]))
    stops = ak.layout.Index64(np.array([3, 3, 6, 8, 9]))
    listarray = ak.layout.ListArray64(starts, stops, content)

    assert ak.to_list(listarray) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
    ]
    value = ak.layout.NumpyArray(np.array([55]))
    assert ak.to_list(listarray.fillna(value)) == [
        [0.0, 1.1, 2.2],
        [],
        [4.4, 5.5],
        [5.5, 6.6, 7.7],
        [8.8],
    ]


def test_fillna_unionarray():
    content1 = ak.from_iter([[], [1.1], [2.2, 2.2]], highlevel=False)
    content2 = ak.from_iter([[2, 2], [1], []], highlevel=False)
    tags = ak.layout.Index8(np.array([0, 1, 0, 1, 0, 1], dtype=np.int8))
    index = ak.layout.Index64(np.array([0, 0, 1, 1, 2, 2], dtype=np.int64))
    array = ak.layout.UnionArray8_64(tags, index, [content1, content2])
    assert ak.to_list(array) == [[], [2, 2], [1.1], [1], [2.2, 2.2], []]

    padded_array = array.rpad(2, 1)
    assert ak.to_list(padded_array) == [
        [None, None],
        [2, 2],
        [1.1, None],
        [1, None],
        [2.2, 2.2],
        [None, None],
    ]
    value = ak.layout.NumpyArray(np.array([777]))
    assert ak.to_list(padded_array.fillna(value)) == [
        [777, 777],
        [2, 2],
        [1.1, 777],
        [1, 777],
        [2.2, 2.2],
        [777, 777],
    ]


def test_highlevel():
    array = ak.Array([[1.1, 2.2, None, 3.3], [], [4.4, None, 5.5]])
    assert ak.to_list(ak.fill_none(array, 999, axis=1)) == [
        [1.1, 2.2, 999, 3.3],
        [],
        [4.4, 999, 5.5],
    ]
    assert ak.to_list(ak.fill_none(array, [1, 2, 3], axis=1)) == [
        [1.1, 2.2, [1, 2, 3], 3.3],
        [],
        [4.4, [1, 2, 3], 5.5],
    ]
    assert ak.to_list(ak.fill_none(array, [], axis=1)) == [
        [1.1, 2.2, [], 3.3],
        [],
        [4.4, [], 5.5],
    ]
    assert ak.to_list(ak.fill_none(array, {"x": 999}, axis=1)) == [
        [1.1, 2.2, {"x": 999}, 3.3],
        [],
        [4.4, {"x": 999}, 5.5],
    ]

    array = ak.Array([[1.1, 2.2, 3.3], None, [], None, [4.4, 5.5]])
    assert ak.to_list(ak.fill_none(array, 999, axis=0)) == [
        [1.1, 2.2, 3.3],
        999,
        [],
        999,
        [4.4, 5.5],
    ]
    assert ak.to_list(ak.fill_none(array, [1, 2, 3], axis=0)) == [
        [1.1, 2.2, 3.3],
        [1, 2, 3],
        [],
        [1, 2, 3],
        [4.4, 5.5],
    ]
    assert ak.to_list(ak.fill_none(array, {"x": 999}, axis=0)) == [
        [1.1, 2.2, 3.3],
        {"x": 999},
        [],
        {"x": 999},
        [4.4, 5.5],
    ]
    assert ak.to_list(ak.fill_none(array, [], axis=0)) == [
        [1.1, 2.2, 3.3],
        [],
        [],
        [],
        [4.4, 5.5],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


class Dummy(ak.Record):
    @property
    def broken(self):
        raise AttributeError("I'm broken!")


def test():
    behavior = {}
    behavior["Dummy"] = Dummy

    recordarray = ak.from_iter([{"x": 1}, {"x": 2}, {"x": 3}], highlevel=False)
    recordarray.setparameter("__record__", "Dummy")
    array = ak.Array(recordarray, behavior=behavior)

    with pytest.raises(AttributeError) as err:
        array[1].broken
    assert str(err.value) == "I'm broken!"  # not "no field named 'broken'"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_na_union():
    one = ak.from_iter([1, None, 3], highlevel=False)
    two = ak.from_iter([[], [1], None, [3, 3, 3]], highlevel=False)
    tags = ak.layout.Index8(np.array([0, 1, 1, 0, 0, 1, 1], dtype=np.int8))
    index = ak.layout.Index64(np.array([0, 0, 1, 1, 2, 2, 3], dtype=np.int64))
    array = ak.Array(
        ak.layout.UnionArray8_64(tags, index, [one, two]), check_valid=True
    )
    assert ak.to_list(array) == [1, [], [1], None, 3, None, [3, 3, 3]]

    assert ak.to_list(ak.is_none(array)) == [
        False,
        False,
        False,
        True,
        False,
        True,
        False,
    ]


class DummyRecord(ak.Record):
    def __repr__(self):
        return "<{0}>".format(self.x)


class DummyArray(ak.Array):
    def __repr__(self):
        return "<DummyArray {0}>".format(" ".join(repr(x) for x in self))


class DeepDummyArray(ak.Array):
    def __repr__(self):
        return "<DeepDummyArray {0}>".format(" ".join(repr(x) for x in self))


def test_behaviors():
    behavior = {}
    behavior["Dummy"] = DummyRecord
    behavior[".", "Dummy"] = DummyArray
    behavior["*", "Dummy"] = DeepDummyArray

    content = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    recordarray = ak.layout.RecordArray({"x": content})
    recordarray.setparameter("__record__", "Dummy")

    array = ak.Array(recordarray, behavior=behavior, check_valid=True)
    assert repr(array) == "<DummyArray <1.1> <2.2> <3.3> <4.4> <5.5>>"
    assert repr(array[0]) == "<1.1>"

    offsets = ak.layout.Index64(np.array([0, 3, 3, 5], dtype=np.int64))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, recordarray)

    array2 = ak.Array(listoffsetarray, behavior=behavior, check_valid=True)

    assert array2.layout.parameter("__record__") is None
    assert array2.layout.purelist_parameter("__record__") == "Dummy"

    assert (
        repr(array2)
        == "<DeepDummyArray <DummyArray <1.1> <2.2> <3.3>> <DummyArray > <DummyArray <4.4> <5.5>>>"
    )
    assert repr(array2[0]) == "<DummyArray <1.1> <2.2> <3.3>>"
    assert repr(array2[0, 0]) == "<1.1>"

    recordarray2 = ak.layout.RecordArray({"outer": listoffsetarray})

    array3 = ak.Array(recordarray2, behavior=behavior, check_valid=True)
    assert type(array3) is ak.Array
    assert type(array3["outer"]) is DeepDummyArray
    assert (
        repr(array3["outer"])
        == "<DeepDummyArray <DummyArray <1.1> <2.2> <3.3>> <DummyArray > <DummyArray <4.4> <5.5>>>"
    )


def test_flatten():
    assert ak.to_list(
        ak.flatten(
            ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True), axis=1
        )
    ) == [1.1, 2.2, 3.3, 4.4, 5.5]


def test_string_equal():
    trials = [
        (
            ["one", "two", "", "three", "four", "", "five", "six", ""],
            ["one", "two", "", "three", "four", "", "five", "six", ""],
        ),
        (
            ["one", "two", "", "three", "four", "", "five", "six"],
            ["one", "two", "", "three", "four", "", "five", "six"],
        ),
        (
            ["one", "two", "", "three", "four", "", "five", "six", ""],
            ["one", "Two", "", "threE", "four", "", "five", "siX", ""],
        ),
        (
            ["one", "two", "", "three", "four", "", "five", "six"],
            ["one", "Two", "", "threE", "four", "", "five", "siX"],
        ),
        (
            ["one", "two", "", "thre", "four", "", "five", "six", ""],
            ["one", "two", "", "three", "four", "", "five", "six", ""],
        ),
        (
            ["one", "two", "", "thre", "four", "", "five", "six"],
            ["one", "two", "", "three", "four", "", "five", "six"],
        ),
        (
            ["one", "two", "", "three", "four", "", "five", "six", ""],
            ["one", "two", ":)", "three", "four", "", "five", "six", ""],
        ),
        (
            ["one", "two", "", "three", "four", "", "five", "six"],
            ["one", "two", ":)", "three", "four", "", "five", "six"],
        ),
        (
            ["one", "two", "", "three", "four", "", "five", "six", ""],
            ["", "two", "", "three", "four", "", "five", "six", ""],
        ),
        (
            ["one", "two", "", "three", "four", "", "five", "six"],
            ["", "two", "", "three", "four", "", "five", "six"],
        ),
    ]

    for left, right in trials:
        assert ak.to_list(
            ak.Array(left, check_valid=True) == ak.Array(right, check_valid=True)
        ) == [x == y for x, y in zip(left, right)]


def test_string_equal2():
    assert ak.to_list(
        ak.Array(["one", "two", "three", "two", "two", "one"], check_valid=True)
        == "two"
    ) == [False, True, False, True, True, False]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_from_iter():
    assert ak.from_iter([1 + 1j, 2 + 2j, 3 + 3j]).tolist() == [1 + 1j, 2 + 2j, 3 + 3j]
    assert ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]).tolist() == [
        [1 + 1j, 2 + 2j],
        [],
        [3 + 3j],
    ]

    # First encounter of a complex number should promote previous integers and
    # reals into complex numbers:
    assert ak.from_iter([1, 2.2, 3 + 3j]).tolist() == [1.0 + 0j, 2.2 + 0j, 3.0 + 3j]
    assert ak.from_iter([1, 3 + 3j]).tolist() == [1.0 + 0j, 3.0 + 3j]

    # Just as the first encounter of a real number promotes previous integers
    # into reals:
    assert str(ak.from_iter([1, 2.2]).type) == "2 * float64"
    assert ak.from_iter([1, 2.2]).tolist() == [1.0, 2.2]
    builder = ak.ArrayBuilder()
    assert str(ak.type(builder)) == "0 * unknown"
    builder.integer(1)
    assert str(ak.type(builder)) == "1 * int64"
    builder.real(2.2)
    assert str(ak.type(builder)) == "2 * float64"

    # For that matter, ArrayBuilder is missing a high-level interface to complex:
    builder.complex(3 + 3j)
    assert str(ak.type(builder)) == "3 * complex128"


def test_from_json():
    array = ak.from_json('[{"r": 1.1, "i": 1.0}, {"r": 2.2, "i": 2.0}]')
    assert array.tolist() == [
        {"r": 1.1, "i": 1.0},
        {"r": 2.2, "i": 2.0},
    ]
    array = ak.from_json(
        '[{"r": 1.1, "i": 1.0}, {"r": 2.2, "i": 2.0}]', complex_record_fields=("r", "i")
    )
    assert array.tolist() == [(1.1 + 1j), (2.2 + 2j)]

    # Somewhere in from_json, a handler that turns integer record fields into
    # parts of a complex number is missing.
    array = ak.from_json(
        '[{"r": 1, "i": 1}, {"r": 2, "i": 2}]', complex_record_fields=("r", "i")
    )
    assert array.tolist() == [(1 + 1j), (2 + 2j)]

    # This should fail with some message like "complex number fields must be numbers,"
    # not "called 'end_record' without 'begin_record' at the same level before it."
    with pytest.raises(ValueError) as err:
        array = ak.from_json(
            '[{"r": [], "i": 1}, {"r": [1, 2], "i": 2}]',
            complex_record_fields=("r", "i"),
        )
        assert array["r"].type == array["i"].type
    assert "Complex number fields must be numbers" in str(err)

    # These shouldn't be recognized as complex number records because they have
    # only one of the two fields.
    assert ak.from_json(
        '[{"r": 1}, {"r": 2}]', complex_record_fields=("r", "i")
    ).tolist() == [{"r": 1}, {"r": 2}]
    assert ak.from_json(
        '[{"i": 1}, {"i": 2}]', complex_record_fields=("r", "i")
    ).tolist() == [{"i": 1}, {"i": 2}]
    assert ak.from_json(
        '[{"r": 1.1}, {"r": 2.2}]', complex_record_fields=("r", "i")
    ).tolist() == [{"r": 1.1}, {"r": 2.2}]
    assert ak.from_json(
        '[{"i": 1.1}, {"i": 2.2}]', complex_record_fields=("r", "i")
    ).tolist() == [{"i": 1.1}, {"i": 2.2}]

    # In this one, the extra field should simply be ignored. A record with *at least*
    # the two specified fields should be recognized as a complex number, so that
    # the protocol can include a type marker, as some protocols do.
    array = ak.from_json(
        '[{"r": 1.1, "i": 1.0, "another": []}, {"r": 2.2, "i": 2.0, "another": [1, 2, 3]}]',
        complex_record_fields=("r", "i"),
    )
    assert array.tolist() == [(1.1 + 1j), (2.2 + 2j)]


def test_to_json():
    # Complex numbers can't be converted to JSON without setting 'complex_record_fields',
    # but the error messages should refer to that name now. (I changed the name at
    # high-level, but not in the error messages emitted by C++ code.)
    with pytest.raises(ValueError) as err:
        ak.to_json(ak.from_iter([1 + 1j, 2 + 2j, 3 + 3j]))
    assert "needs both" not in str(err)

    expectation = [{"r": 1.0, "i": 1.0}, {"r": 2.0, "i": 2.0}, {"r": 3.0, "i": 3.0}]
    assert expectation == json.loads(
        ak.to_json(
            ak.from_iter([1 + 1j, 2 + 2j, 3 + 3j]), complex_record_fields=("r", "i")
        )
    )
    expectation = [
        [{"r": 1.0, "i": 1.0}, {"r": 2.0, "i": 2.0}],
        [],
        [{"r": 3.0, "i": 3.0}],
    ]
    assert expectation == json.loads(
        ak.to_json(
            ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]),
            complex_record_fields=("r", "i"),
        )
    )


def test_reducers():
    # axis=None reducers are implemented in NumPy.
    assert ak.sum(ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]])) == 6 + 6j
    assert ak.prod(ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]])) == -12 + 12j

    # axis != None reducers are implemented in libawkward; this should be ReducerSum.
    assert ak.sum(ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]), axis=1).tolist() == [
        3 + 3j,
        0 + 0j,
        3 + 3j,
    ]
    # And this is in ReducerProd.
    assert ak.prod(ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]), axis=1).tolist() == [
        0 + 4j,
        1 + 0j,
        3 + 3j,
    ]

    # ReducerCount, ReducerCountNonzero, ReducerAny, and ReducerAll work.
    assert ak.count(
        ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]), axis=1
    ).tolist() == [2, 0, 1]
    assert ak.count_nonzero(
        ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]), axis=1
    ).tolist() == [2, 0, 1]
    assert ak.any(ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]), axis=1).tolist() == [
        True,
        False,
        True,
    ]
    assert ak.all(ak.from_iter([[1 + 1j, 2 + 2j], [], [3 + 3j]]), axis=1).tolist() == [
        True,
        True,
        True,
    ]
    assert ak.any(
        ak.from_iter([[1 + 1j, 2 + 2j, 0 + 0j], [], [3 + 3j]]), axis=1
    ).tolist() == [True, False, True]
    assert ak.all(
        ak.from_iter([[1 + 1j, 2 + 2j, 0 + 0j], [], [3 + 3j]]), axis=1
    ).tolist() == [False, True, True]


def test_minmax():
    assert ak.min(ak.from_iter([[1 + 5j, 2 + 4j], [], [3 + 3j]])) == 1 + 5j
    assert ak.max(ak.from_iter([[1 + 5j, 2 + 4j], [], [3 + 3j]])) == 3 + 3j

    assert ak.min(ak.from_iter([[1 + 5j, 2 + 4j], [], [3 + 3j]]), axis=1).tolist() == [
        1 + 5j,
        None,
        3 + 3j,
    ]
    assert ak.max(ak.from_iter([[1 + 5j, 2 + 4j], [], [3 + 3j]]), axis=1).tolist() == [
        2 + 4j,
        None,
        3 + 3j,
    ]

    assert ak.argmin(
        ak.from_iter([[1 + 5j, 2 + 4j], [], [3 + 3j]]), axis=1
    ).tolist() == [0, None, 0]
    assert ak.argmax(
        ak.from_iter([[1 + 5j, 2 + 4j], [], [3 + 3j]]), axis=1
    ).tolist() == [1, None, 0]


@pytest.mark.skip(reason="Remember to implement sorting for complex numbers.")
def test_sort():
    assert ak.sort(ak.from_iter([[2 + 4j, 1 + 5j], [], [3 + 3j]])).tolist() == [
        [1 + 5j, 2 + 4j],
        [],
        [3 + 3j],
    ]
    assert ak.argsort(ak.from_iter([[2 + 4j, 1 + 5j], [], [3 + 3j]])).tolist() == [
        [1, 0],
        [],
        [0],
    ]


def test_numpy():
    # This all is fine.
    assert np.array_equal(
        ak.to_numpy(ak.from_iter([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]])),
        np.array([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]]),
    )
    assert (
        str(ak.to_numpy(ak.from_iter([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]])).dtype)
        == "complex128"
    )
    assert ak.Array(np.array([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]])).tolist() == [
        [(1 + 1j), (2 + 2j)],
        [(3 + 3j), (4 + 4j)],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


numba = pytest.importorskip("numba")


def test():
    @numba.njit
    def f1(x):
        return x[0], x[1]

    array = ak.Array(np.arange(4).reshape(2, 2)[:, 0])

    assert f1.py_func(array) == (0, 2)
    assert f1(array) == (0, 2)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_UnmaskedArray():
    content_float64 = ak.layout.NumpyArray(
        np.array([0.25, 0.5, 3.5, 4.5, 5.5], dtype=np.float64)
    )
    array_float64 = ak.layout.UnmaskedArray(content_float64)
    assert ak.to_list(array_float64) == [0.25, 0.5, 3.5, 4.5, 5.5]
    assert str(ak.type(content_float64)) == "float64"
    assert str(ak.type(ak.Array(content_float64))) == "5 * float64"
    assert str(ak.type(array_float64)) == "?float64"
    assert str(ak.type(ak.Array(array_float64))) == "5 * ?float64"

    assert np.can_cast(np.float32, np.float64) is True
    assert np.can_cast(np.float64, np.float32, "unsafe") is True
    assert np.can_cast(np.float64, np.int8, "unsafe") is True

    content_float32 = ak.values_astype(content_float64, "float32", highlevel=False)
    array_float32 = ak.layout.UnmaskedArray(content_float32)
    assert ak.to_list(array_float32) == [0.25, 0.5, 3.5, 4.5, 5.5]
    assert str(ak.type(content_float32)) == "float32"
    assert str(ak.type(ak.Array(content_float32))) == "5 * float32"
    assert str(ak.type(array_float32)) == "?float32"
    assert str(ak.type(ak.Array(array_float32))) == "5 * ?float32"

    content_int8 = ak.values_astype(content_float64, "int8", highlevel=False)
    array_int8 = ak.layout.UnmaskedArray(content_int8)
    assert ak.to_list(array_int8) == [0, 0, 3, 4, 5]
    assert str(ak.type(content_int8)) == "int8"
    assert str(ak.type(ak.Array(content_int8))) == "5 * int8"
    assert str(ak.type(array_int8)) == "?int8"
    assert str(ak.type(ak.Array(array_int8))) == "5 * ?int8"

    content_from_int8 = ak.values_astype(content_int8, "float64", highlevel=False)
    array_from_int8 = ak.layout.UnmaskedArray(content_from_int8)
    assert ak.to_list(array_from_int8) == [0, 0, 3, 4, 5]
    assert str(ak.type(content_from_int8)) == "float64"
    assert str(ak.type(ak.Array(content_from_int8))) == "5 * float64"
    assert str(ak.type(array_from_int8)) == "?float64"
    assert str(ak.type(ak.Array(array_from_int8))) == "5 * ?float64"


def test_RegularArray_and_ListArray():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    regulararray = ak.layout.RegularArray(listoffsetarray, 2, zeros_length=0)
    starts = ak.layout.Index64(np.array([0, 1]))
    stops = ak.layout.Index64(np.array([2, 3]))
    listarray = ak.layout.ListArray64(starts, stops, regulararray)

    assert str(ak.type(content)) == "float64"
    assert str(ak.type(regulararray)) == "2 * var * float64"
    assert str(ak.type(listarray)) == "var * 2 * var * float64"

    regulararray_int8 = ak.values_astype(regulararray, "int8", highlevel=False)
    assert str(ak.type(regulararray_int8)) == "2 * var * int8"

    listarray_bool = ak.values_astype(listarray, "bool", highlevel=False)
    assert str(ak.type(listarray_bool)) == "var * 2 * var * bool"


def test_ufunc_afterward():
    assert (
        ak.values_astype(ak.Array([{"x": 1.1}, {"x": 3.3}]), np.float32)["x"] + 1
    ).tolist() == [2.0999999046325684, 4.300000190734863]


def test_string():
    assert ak.values_astype(
        ak.Array([{"x": 1.1, "y": "hello"}]), np.float32
    ).tolist() == [{"x": 1.100000023841858, "y": "hello"}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import numpy as np
import awkward as ak


def test_NumpyArray():
    array = ak.layout.NumpyArray(
        np.array(["1chchc", "1chchc", "2sss", "3", "4", "5"], dtype=object),
        parameters={"__array__": "categorical"},
    )
    assert ak.is_valid(ak.Array(array)) is False
    # FIXME? assert array.is_unique() is False
    array2 = ak.layout.NumpyArray(np.array([5, 6, 1, 3, 4, 5]))
    assert array2.is_unique() is False


def test_ListOffsetArray():
    array = ak.from_iter(["one", "two", "three", "four", "five"], highlevel=False)
    assert ak.to_list(array.sort(0, True, True)) == [
        "five",
        "four",
        "one",
        "three",
        "two",
    ]
    assert array.is_unique() is True

    array2 = ak.from_iter(["one", "two", "one", "four", "two"], highlevel=False)
    assert ak.to_list(array2.sort(0, True, True)) == [
        "four",
        "one",
        "one",
        "two",
        "two",
    ]
    assert array2.is_unique() is False

    content = ak.layout.NumpyArray(
        np.array([3.3, 1.1, 2.2, 0.0, 4.4, 9.9, 6.6, 7.7, 8.8, 5.5])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    assert ak.to_list(listoffsetarray) == [
        [3.3, 1.1, 2.2],
        [],
        [0.0, 4.4],
        [9.9],
        [6.6, 7.7, 8.8, 5.5],
        [],
    ]
    assert listoffsetarray.is_unique() is True

    content = ak.layout.NumpyArray(
        np.array([3.3, 1.1, 2.2, 0.0, 4.4, 9.9, 2.2, 3.3, 1.1, 5.5])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    assert ak.to_list(listoffsetarray) == [
        [3.3, 1.1, 2.2],
        [],
        [0.0, 4.4],
        [9.9],
        [2.2, 3.3, 1.1],
        [5.5],
    ]
    assert listoffsetarray.is_unique() is False

    content2 = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 1.1, 5.5, 6.6, 7.7, 2.2, 9.9])
    )
    offsets2 = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray2 = ak.layout.ListOffsetArray64(offsets2, content2)
    assert ak.to_list(listoffsetarray2) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 1.1],
        [5.5],
        [6.6, 7.7, 2.2, 9.9],
        [],
    ]
    assert listoffsetarray2.is_unique() is True


def test_RegularArray():
    content = ak.layout.NumpyArray(
        np.array(
            [
                2.1,
                8.4,
                7.4,
                1.6,
                2.2,
                3.4,
                6.2,
                5.4,
                1.5,
                3.9,
                3.8,
                3.0,
                8.5,
                6.9,
                4.3,
                3.6,
                6.7,
                1.8,
                3.2,
            ]
        )
    )
    index = ak.layout.Index64(
        np.array([13, 9, 13, 4, 8, 3, 15, -1, 16, 2, 8], dtype=np.int64)
    )
    indexedarray = ak.layout.IndexedOptionArray64(index, content)
    regular_array = ak.layout.RegularArray(indexedarray, 3, zeros_length=0)
    assert ak.to_list(regular_array) == [
        [6.9, 3.9, 6.9],
        [2.2, 1.5, 1.6],
        [3.6, None, 6.7],
    ]
    assert regular_array.is_unique() is True

    index2 = ak.layout.Index64(
        np.array([13, 9, 13, 9, 13, 13, -1, -1, -1, 2, 8], dtype=np.int64)
    )
    indexedarray2 = ak.layout.IndexedOptionArray64(index2, content)
    regular_array2 = ak.layout.RegularArray(indexedarray2, 3, zeros_length=0)
    assert ak.to_list(regular_array2) == [
        [6.9, 3.9, 6.9],
        [3.9, 6.9, 6.9],
        [None, None, None],
    ]
    assert regular_array2.is_unique() is False


def test_IndexedArray():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index = ak.layout.Index64(np.array([0, 2, 4, 6, 8, 9], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray64(index, content)

    assert ak.to_list(indexedarray) == [0.0, 2.2, 4.4, 6.6, 8.8, 9.9]
    assert indexedarray.is_unique() is True

    index2 = ak.layout.Index64(np.array([0, 2, 9, 6, 0, 9], dtype=np.int64))
    indexedarray2 = ak.layout.IndexedArray64(index2, content)

    assert ak.to_list(indexedarray2) == [0.0, 2.2, 9.9, 6.6, 0.0, 9.9]
    assert indexedarray2.is_unique() is False


def test_RecordArray():
    array = ak.Array(
        [
            {"x": 0.0, "y": []},
            {"x": 8.0, "y": [1]},
            {"x": 2.2, "y": [2, 2]},
            {"x": 3.3, "y": [3, 1, 3]},
            {"x": 4.4, "y": [4, 1, 1, 4]},
            {"x": 5.5, "y": [5, 4, 5]},
            {"x": 1.1, "y": [6, 1]},
            {"x": 7.7, "y": [7]},
            {"x": 0.0, "y": []},
        ]
    )

    assert array.layout.is_unique() is True
    assert array.x.layout.is_unique() is False
    assert array.y.layout.is_unique() is True


def test_same_categories():
    categories = ak.Array(["one", "two", "three"])
    index1 = ak.layout.Index64(np.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.layout.Index64(np.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.layout.IndexedArray64(
        index1, categories.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.layout.IndexedArray64(
        index2, categories.layout, parameters={"__array__": "categorical"}
    )
    assert ak.to_list(categorical1.sort(0, True, True)) == [
        "one",
        "one",
        "one",
        "three",
        "three",
        "three",
        "two",
        "two",
    ]
    assert categorical1.is_unique() is False
    assert categorical1.content.is_unique() is True
    assert categorical2.is_unique() is False

    array1 = ak.Array(categorical1)
    assert array1.tolist() == [
        "one",
        "three",
        "three",
        "two",
        "three",
        "one",
        "two",
        "one",
    ]

    array2 = ak.Array(categorical2)
    assert array2.tolist() == ["two", "two", "three", "two", "one", "one", "one", "two"]

    assert (array1 == array2).tolist() == [
        False,
        False,
        True,
        True,
        False,
        True,
        False,
        False,
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    assert ak.unflatten(array, 5).tolist() == [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]
    assert ak.unflatten(array, [3, 0, 2, 1, 4]).tolist() == [
        [0, 1, 2],
        [],
        [3, 4],
        [5],
        [6, 7, 8, 9],
    ]
    assert ak.unflatten(array, [3, None, 2, 1, 4]).tolist() == [
        [0, 1, 2],
        None,
        [3, 4],
        [5],
        [6, 7, 8, 9],
    ]

    original = ak.Array([[0, 1, 2], [], [3, 4], [5], [6, 7, 8, 9]])
    counts = ak.num(original)
    array = ak.flatten(original)
    assert counts.tolist() == [3, 0, 2, 1, 4]
    assert array.tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    assert ak.unflatten(array, counts).tolist() == [
        [0, 1, 2],
        [],
        [3, 4],
        [5],
        [6, 7, 8, 9],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    data = ak.Array([[1, 3, 5, 4, 2], [], [2, 3, 1], [5]])
    assert ak.min(data, axis=1, initial=4).tolist() == [1, None, 1, 4]
    assert ak.max(data, axis=1, initial=4).tolist() == [5, None, 4, 5]

    data = ak.Array([[1.1, 3.3, 5.5, 4.4, 2.2], [], [2.2, 3.3, 1.1], [5.5]])
    assert ak.min(data, axis=1, initial=4).tolist() == [1.1, None, 1.1, 4]
    assert ak.max(data, axis=1, initial=4).tolist() == [5.5, None, 4, 5.5]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert (
        ak.Array([[1, 2], [3, 4]])[[True, False], 0].ndim
        == np.array([[1, 2], [3, 4]])[[True, False], 0].ndim
    )
    assert (
        ak.Array([[1, 2], [3, 4]])[[False, False], 0].ndim
        == np.array([[1, 2], [3, 4]])[[False, False], 0].ndim
    )

    assert (
        ak.Array([[1, 2], [3, 4]])[[True, False], 0].ndim
        == np.array([[1, 2], [3, 4]])[[True, False]][:, 0].ndim
    )
    assert (
        ak.Array([[1, 2], [3, 4]])[[False, False], 0].ndim
        == np.array([[1, 2], [3, 4]])[[False, False]][:, 0].ndim
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_partition_methods():
    one, two, three = ak.Array([1, 2, 3]), ak.Array([4, 5]), ak.Array([6])
    no = one.layout
    yes = ak.partitioned([one, two, three]).layout

    ONE, TWO, THREE = ak.Array([[1, 2], [3]]), ak.Array([[4, 5]]), ak.Array([[6]])
    NO = ONE.layout
    YES = ak.partitioned([ONE, TWO, THREE]).layout

    assert no.withparameter("hey", "you").parameter("hey") == "you"
    assert yes.withparameter("hey", "you").parameter("hey") == "you"
    assert no.type({}) == yes.type({})
    assert no.form == yes.form
    assert len(no) == 3
    assert len(yes) == 6
    assert list(no) == [1, 2, 3]
    assert list(yes) == [1, 2, 3, 4, 5, 6]
    assert no.kernels == yes.kernels == "cpu"
    assert no.caches == yes.caches
    assert no.tojson() == "[1,2,3]"
    assert yes.tojson() == "[1,2,3,4,5,6]"
    assert no.nbytes < yes.nbytes
    assert no.deep_copy().tojson() == "[1,2,3]"
    assert yes.deep_copy().tojson() == "[1,2,3,4,5,6]"
    assert no.numfields == yes.numfields
    assert no.haskey("doesn't") == yes.haskey("doesn't")
    assert no.keys() == yes.keys()
    assert no.purelist_isregular == yes.purelist_isregular
    assert no.purelist_depth == yes.purelist_depth
    assert no.branch_depth == yes.branch_depth
    assert no.minmax_depth == yes.minmax_depth
    assert list(no.getitem_nothing()) == list(yes.getitem_nothing())
    assert no.getitem_at_nowrap(2) == yes.getitem_at_nowrap(2)
    assert list(no.getitem_range_nowrap(1, 2)) == list(yes.getitem_range_nowrap(1, 2))
    assert no.validityerror() is None
    assert yes.validityerror() is None
    assert list(no.fillna(ak.to_layout([123]))) == [1, 2, 3]
    assert list(yes.fillna(ak.to_layout([123]))) == [1, 2, 3, 4, 5, 6]
    assert no.num(0) == 3
    assert yes.num(0) == 6
    assert list(NO.flatten(1)) == [1, 2, 3]
    assert list(YES.flatten(1)) == [1, 2, 3, 4, 5, 6]
    assert [np.asarray(x).tolist() for x in NO.offsets_and_flatten(1)] == [
        [0, 2, 3],
        [1, 2, 3],
    ]
    assert [np.asarray(x).tolist() for x in YES.offsets_and_flatten(1)] == [
        [0, 2, 3, 5, 6],
        [1, 2, 3, 4, 5, 6],
    ]
    assert ak.to_list(NO.rpad(3, 1)) == [[1, 2, None], [3, None, None]]
    assert ak.to_list(YES.rpad(3, 1)) == [
        [1, 2, None],
        [3, None, None],
        [4, 5, None],
        [6, None, None],
    ]
    assert ak.to_list(NO.rpad_and_clip(3, 1)) == [[1, 2, None], [3, None, None]]
    assert ak.to_list(YES.rpad_and_clip(3, 1)) == [
        [1, 2, None],
        [3, None, None],
        [4, 5, None],
        [6, None, None],
    ]
    assert no.mergeable(no)
    assert yes.mergeable(yes)
    assert ak.to_list(no.merge(no)) == [1, 2, 3, 1, 2, 3]
    assert ak.to_list(yes.merge(yes)) == [1, 2, 3, 4, 5, 6, 1, 2, 3, 4, 5, 6]
    assert ak.to_list(no.mergemany([no])) == [1, 2, 3, 1, 2, 3]
    assert ak.to_list(yes.mergemany([yes])) == [1, 2, 3, 4, 5, 6, 1, 2, 3, 4, 5, 6]
    assert NO.axis_wrap_if_negative(-1) == 1
    assert YES.axis_wrap_if_negative(-1) == 1
    assert no.count(0, False, False) == 3
    assert yes.count(0, False, False) == 6
    assert no.count_nonzero(0, False, False) == 3
    assert yes.count_nonzero(0, False, False) == 6
    assert no.sum(0, False, False) == 1 + 2 + 3
    assert yes.sum(0, False, False) == 1 + 2 + 3 + 4 + 5 + 6
    assert no.prod(0, False, False) == 1 * 2 * 3
    assert yes.prod(0, False, False) == 1 * 2 * 3 * 4 * 5 * 6
    assert no.any(0, False, False) is True
    assert yes.any(0, False, False) is True
    assert no.all(0, False, False) is True
    assert yes.all(0, False, False) is True
    assert no.min(0, False, False, None) == 1
    assert yes.min(0, False, False, None) == 1
    assert no.max(0, False, False, None) == 3
    assert yes.max(0, False, False, None) == 6
    assert no.argmin(0, False, False) == 0
    assert yes.argmin(0, False, False) == 0
    assert no.argmax(0, False, False) == 2
    assert yes.argmax(0, False, False) == 5
    assert ak.to_list(NO.localindex(1)) == [[0, 1], [0]]
    assert ak.to_list(YES.localindex(1)) == [[0, 1], [0], [0, 1], [0]]
    assert ak.to_list(NO.combinations(2, True, None, {}, 1)) == [
        [(1, 1), (1, 2), (2, 2)],
        [(3, 3)],
    ]
    assert ak.to_list(YES.combinations(2, True, None, {}, 1)) == [
        [(1, 1), (1, 2), (2, 2)],
        [(3, 3)],
        [(4, 4), (4, 5), (5, 5)],
        [(6, 6)],
    ]
    assert ak.to_list(no.sort(0, True, True)) == [1, 2, 3]
    assert ak.to_list(NO.sort(1, True, True)) == [[1, 2], [3]]
    assert ak.to_list(yes.sort(0, True, True)) == [1, 2, 3, 4, 5, 6]
    assert ak.to_list(YES.sort(0, True, True)) == [[1, 2], [3], [4, 5], [6]]
    assert ak.to_list(no.argsort(0, True, True)) == [0, 1, 2]
    assert ak.to_list(NO.argsort(1, True, True)) == [[0, 1], [0]]
    assert ak.to_list(yes.argsort(0, True, True)) == [0, 1, 2, 3, 4, 5]
    assert ak.to_list(YES.argsort(1, True, True)) == [[0, 1], [0], [0, 1], [0]]
    assert np.asarray(no.numbers_to_type("float64")).dtype == np.dtype("f8")
    assert np.asarray(yes.numbers_to_type("float64")).dtype == np.dtype("f8")
    assert no.is_unique() is True
    assert NO.is_unique() is True
    assert yes.is_unique() is True
    assert YES.is_unique() is True
    assert ak.to_list(no.copy_to("cpu")) == [1, 2, 3]
    assert ak.to_list(yes.copy_to("cpu")) == [1, 2, 3, 4, 5, 6]
    assert ak.to_list(
        no.carry(ak.layout.Index64(np.array([2, 1, 0], np.int64)), False)
    ) == [3, 2, 1]
    assert ak.to_list(
        yes.carry(ak.layout.Index64(np.array([5, 4, 3, 2, 1, 0], np.int64)), False)
    ) == [6, 5, 4, 3, 2, 1]
